import os
import sys
try:
    import yaml
except ImportError:
    pass
import argparse
from gpstime import gpstime
import numpy as np
from PyQt4 import QtGui
import signal
try:
    from setproctitle import setproctitle
except ImportError:
    def setproctitle(*args):
        pass
import logging

from . import __version__
from . import util
from .main import set_background, NDScope

logging.addLevelName(5, 'DATA')
logging.basicConfig(
    level=os.getenv('LOG_LEVEL', 'WARNING'),
    format="%(threadName)s:%(message)s",
)

##################################################

DEFAULT_TIME_WINDOW_ONLINE = 2
DEFAULT_TIME_WINDOW_OFFLINE = 10


PROG = 'ndscope'
DESCRIPTION = 'Next generation NDS oscilloscope'
USAGE = '''

If no time is specified, online data will be plotted.  The -t option
may be used to specify a time in the past.  Multiple -t options
specify a time range.  Times can be expressed in GPS, Gregorian, or
relative (e.g.  "3 days ago").  Remember to use quotes around
command-line time specifications that include spaces.  The -w option
allows specifying an initial time window when providing a single -t
option, or an initial lookback time window for online mode.

Left and right mouse buttons control pan and zoom respectively, as
does center wheel.  Missing data will automatically be fetched as
needed.  In offline mode, second and minute trend data are substituted
automatically depending on the time scale being requested/displayed.

WARNING: Online mode uses only raw, full-rate data, so large windows
can severely affect performance when working with fast channels.

By default all channels are placed into a grid (row-major ordering).
If the --table/-l option is provided, periods and commas in the
channel list (space separated) will cause new subplots to be created,
with commas starting new rows of plots.  For instance, specifying the
channel list as:

  FOO . BAR BAZ , QUX

will cause three subplots to be created, two in the top row (the
second with two channels BAR and BAZ), and one in the second.  If the
--table option is not provided, periods and commas in the channel list
will be ignored.

Plot configuration can be loaded from ndscope .yaml or StripTool .stp
description files.  An ndscope .yaml template can be generated with
the --gen-template option.  The "plots:" block in the description is a
list of subplot definitions.  Each subplot should include a "channels"
mapping of channel names to curve properties, such as "color",
"scale", and "offset".

NDS server[:port] should be provided in the NDSSERVER environment
variable.

Please report issues: https://git.ligo.org/cds/ndscope/issues

'''


class TimeParseAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=False):
        if getattr(namespace, self.dest) is None:
            setattr(namespace, self.dest, [gpstime.parse(values)])
        else:
            getattr(namespace, self.dest).append(gpstime.parse(values))


parser = argparse.ArgumentParser(
    prog=PROG,
    description=DESCRIPTION+USAGE,
    formatter_class=argparse.RawDescriptionHelpFormatter,
)

parser.add_argument('channels', nargs='+',
                    help="channels, or yaml layout")
parser.add_argument('-t', '--time', action=TimeParseAction, default=[],
                    help="time boundary, in GPS, or natural language (may specify up to two)")
parser.add_argument('-w', '--time-window', type=float, metavar="SECONDS",
                    help="time window in seconds [online: {}s, offline: {}s]".format(
                        DEFAULT_TIME_WINDOW_ONLINE,
                        DEFAULT_TIME_WINDOW_OFFLINE))
# parser.add_argument('-r', '--replay', action='store_true',
#                     help="online replay mode")
lgroup = parser.add_mutually_exclusive_group()
lgroup.add_argument('-g', '--grid', dest='layout', action='store_const', const='grid',
                    help="arrange channels in a grid of plots (default)")
lgroup.add_argument('-s', '--stack', dest='layout', action='store_const', const='stack',
                    help="arrange channels in a vertical stack of plots")
lgroup.add_argument('-l', '--table', dest='layout', action='store_const', const='table',
                    help="subplot table layout (periods start new colums, commas start new rows)")
parser.add_argument('--colspan', action='store_true',
                    help="expand subplots to fill empty columns (EXPERIMENTAL AND BUGGY)")
parser.add_argument('--window-title', metavar='TITLE', default=None,
                    help="application window title")
parser.add_argument('-bw', '--black-on-white', action='store_true', default=None,
                    help="black-on-white plots, instead of white-on-black")
parser.add_argument('--gen-template', action='store_true',
                    help="generate YAML template for layout and dump to stdout")
parser.add_argument('--version', action='version',
                    version='%(prog)s {}'.format(__version__))


def _resolve_template_args(args, template):
    # prefer command line arguments over template definitions
    for arg, val in args.__dict__.items():
        targ = arg.replace('_', '-')
        if val is None and targ in template:
                args.__dict__[arg] = template[targ]


def main():
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    setproctitle(' '.join([PROG] + sys.argv[1:]))

    args = parser.parse_args()
    logging.debug(args)

    if os.path.exists(args.channels[0]):
        tpath = args.channels[0]
        if len(args.channels) > 1:
            parser.error("Only one argument expected when specifying template file")

        if os.path.splitext(tpath)[1] == '.stp':
            template = util.template_from_stp(tpath)
        else:
            try:
                with open(tpath, 'r') as f:
                    template = yaml.load(f)
            except NameError:
                parser.error("YAML descriptions not supported on this system.")

        _resolve_template_args(args, template)
        layout = template['plots']
        if not args.window_title:
            args.window_title = os.path.basename(tpath)

        # extract all channels into a list
        channels = []
        try:
            for plot in layout:
                for chan in plot['channels'].keys():
                    if chan not in channels:
                        channels.append(chan)
        except:
            print("Error parsing description file; please check formatting:")
            raise

    else:
        layout = None
        if args.layout is None:
            args.layout = 'grid'
        channels = [chan for chan in args.channels if chan not in ('.', ',')]
        if not args.window_title:
            args.window_title = ' '.join(channels)

    for arg, val in args.__dict__.items():
        if val is None:
            args.__dict__[arg] = False

    if len(args.time) == 0:
        online = True
        start = None
        end = None
        if not args.time_window:
            args.time_window = DEFAULT_TIME_WINDOW_ONLINE
    elif len(args.time) == 1:
        online = False
        if not args.time_window:
            args.time_window = DEFAULT_TIME_WINDOW_OFFLINE
        start = args.time[0].gps() - args.time_window/2
        end = start + args.time_window
    elif len(args.time) == 2:
        online = False
        if args.time[0] > args.time[1]:
            start = args.time[1].gps()
            end = args.time[0].gps()
        else:
            start = args.time[0].gps()
            end = args.time[1].gps()
        args.time_window = end - start
    else:
        parser.error("May only specify one or two times.")

    args.time_window = abs(args.time_window)
    start_end = (start, end)

    # if args.replay:
    #     if not start:
    #         sys.exit("Must specify times for replay.")
    #     replay = start_end
    # else:
    #     replay = None
    replay = None

    if not layout or args.layout in ['grid', 'stack']:
        # the below parses the channels list arguments into a list of
        # rows, list of columns, list of subplot channels
        # FIXME: it all seems overly complicated

        if args.layout == 'grid':
            args.colspan = False
            num = len(channels)
            rows = int(np.ceil(np.sqrt(num)))
            cols = int(np.ceil(float(num)/rows))
            chan_layout = []
            for r in range(rows):
                chan_layout.append([[]]*(cols))
            r = 0
            c = 0
            for i, chan in enumerate(channels):
                chan_layout[r][c] = [chan]
                c += 1
                if c == cols:
                    c = 0
                    r += 1

        elif args.layout == 'stack':
            chan_layout = []
            for chan in channels:
                chan_layout.append([[chan]])

        elif args.layout == 'table':
            chan_layout = [[[]]]
            r = 0
            c = 0
            for chan in args.channels:
                if chan == '.':
                    chan_layout[r].append([])
                    c += 1
                    continue
                if chan == ',':
                    chan_layout.append([])
                    r += 1
                    chan_layout[r].append([])
                    c = 0
                    continue
                chan_layout[r][c].append(chan)

        else:
            parser.error("Unknown layout option: {}".format(args.layout))

        # construct the layout structure as a list of subplot
        # dictionary definitions

        # determine the total number of rows and columns
        cols = 0
        for row in chan_layout:
            cols = max(len(row), cols)
        rows = len(chan_layout)

        # now make layout, optionally expanding plots to fill multiple
        # columns
        _layout = []
        for r in range(rows):
            row = chan_layout[r]
            sp = None
            cs = 1
            for c in range(cols):
                try:
                    chans = row[c]
                except IndexError:
                    if args.colspan:
                        cs += 1
                    continue
                if chans:
                    if sp:
                        _layout.append(sp + (cs,))
                    sp = (r, c, chans)
                    cs = 1
                elif args.colspan:
                    cs += 1
            if sp:
                _layout.append(sp + (cs,))

        layout = []
        for sp in _layout:
            subplot = {
                'row': sp[0],
                'col': sp[1],
                'colspan': sp[3],
            }
            spchans = {}
            for chan in sp[2]:
                spchans[chan] = {
                    'color': [int(i) for i in list(np.random.rand(3)*255)],
                    'scale': 1,
                    'offset': 0,
                }
            subplot['channels'] = spchans
            subplot['yrange'] = 'auto'
            layout.append(subplot)


    if args.gen_template:
        template = {
            'version': '{} {}'.format(PROG, __version__),
            'window-title': args.window_title,
            'black-on-white': args.black_on_white,
            'time-window': args.time_window,
            'plots': layout,
        }
        print(yaml.dump(template, default_style=False))
        sys.exit()

    if args.black_on_white:
        set_background('w')

    app = QtGui.QApplication([])
    scope = NDScope(layout, start_end=replay)
    scope.setWindowTitle('{}: {}'.format(PROG, args.window_title))
    scope.show()

    if online:
        scope.start(args.time_window)
    else:
        scope.fetch(start_end)

    sys.exit(app.exec_())


if __name__ == '__main__':
    main()
