# -*- coding: utf-8 -*-
from __future__ import division
import os
import copy
import numpy as np
import pyqtgraph as pg
from PyQt4 import QtGui
from PyQt4.QtGui import QStyle
from PyQt4.QtCore import pyqtSignal, pyqtSlot
from gpstime import gpstime
from dateutil.tz import tzutc, tzlocal
import logging

from .designer import Ui_MainWindow
from .data import DataStore
from .trigger import Trigger
from .cursors import Crosshair, TCursors, YCursors

##################################################
# CONFIG

# 4,915,200 points for 16kHz data
TREND_THRESHOLD_S = 300
#   259,200 points for 3*1Hz data
TREND_THRESHOLD_M = 86400

# date/time formatting for GPS conversion
# FIXME: why does '%c' without explicit TZ give very wrong values??
DATETIME_TZ = tzlocal()
DATETIME_FMT = '%c'
#DATETIME_FMT = '%a %b %d %Y %H:%M:%S %Z'

def set_background(color='k'):
    if color == 'w':
        pg.setConfigOption('background', 'w')
        pg.setConfigOption('foreground', 'k')

if os.getenv('ANTIALIAS'):
    pg.setConfigOption('antialias', True)
    print("Anti-aliasing ENABLED")
# pg.setConfigOption('leftButtonPan', False)
# see also ViewBox.setMouseMode(ViewBox.RectMode)
# file:///usr/share/doc/python-pyqtgraph-doc/html/graphicsItems/viewbox.html#pyqtgraph.ViewBox.setMouseMode

##################################################

class NDScope(QtGui.QMainWindow, Ui_MainWindow):
    t0ChangeEvent = pyqtSignal()

    def __init__(self, layout, start_end=None):
        """initilize NDScope object

        `layout` should be a list of subplot definitions, each being a
        dictionary matching the keyword arguments to add_plot()
        (e.g. row, col, colspan, channels, etc.).

        """
        super(NDScope, self).__init__(None)
        self.setupUi(self)

        ##########

        self.data = DataStore(self.statusBar)
        self.start_end = start_end
        self.online = False

        self.t0 = 0
        self.plots = []
        self.curves = {}
        self.colors = {}

        # Crosshair needs Scope for access to t0
        self.Crosshair = Crosshair(self)
        self.TCursors = TCursors()
        self.YCursors = YCursors()

        ##########
        # create plots

        rows = 0
        for subplot in layout:
            self.add_plot(**subplot)
            rows = max(rows, subplot['row'])

        self.referenceTimeLabel = self.graphView.addLabel(
            "GPS Time",
            size='11pt',
            #bold=True,
            row=rows+1,
            col=0,
            colspan=1000,
        )
        # colspan is an arbitrarily large number just so that the
        # label always spans the full width regardless what it is

        ##########
        # data retrieval

        self.data.new_data.connect(self.update_plots)
        self.data.done.connect(self.done)

        self.startstopButton.clicked.connect(self.startstop)
        self.fetchButton.clicked.connect(self.fetch)

        ##########
        # trigger

        self.trigger = Trigger()
        self.trigger.sigLevelChanged.connect(self.update_trigger_level)
        self.triggerSelect.addItems(self.data.channels)
        self.triggerSelect.currentIndexChanged.connect(self.update_trigger_channel)
        self.triggerLevel.setValidator(QtGui.QDoubleValidator())
        self.triggerLevel.returnPressed.connect(self.set_trigger_level)
        self.triggerResetLevel.clicked.connect(self.reset_trigger_level)
        self.triggerSingle.clicked.connect(self.trigger.set_single)
        self.triggerInvert.clicked.connect(self.trigger.set_invert)
        self.triggerGroup.toggled.connect(self.toggle_trigger)

        ##########
        # cursors

        self.crosshairGroup.toggled.connect(self.toggle_crosshair)

        self.cursorTGroup.toggled.connect(self.toggle_t_cursors)
        self.cursorTReset.clicked.connect(self.TCursors.reset)

        self.cursorYGroup.toggled.connect(self.toggle_y_cursors)
        self.cursorYPlot.addItems(['{},{}'.format(p._row, p._col) for p in self.plots])
        self.cursorYPlot.currentIndexChanged.connect(self.set_y_cursors_to_selected)
        self.cursorYReset.clicked.connect(self.YCursors.reset)

        ##########
        # other widget setup

        self.offlineStartGPS.returnPressed.connect(self.fetch)
        self.offlineStartGPS.setValidator(QtGui.QIntValidator())
        self.offlineStopGPS.returnPressed.connect(self.fetch)
        self.offlineStopGPS.setValidator(QtGui.QIntValidator())

        ##########

        self.plot0.sigXRangeChanged.connect(self.update_range)

    ##########

    def add_plot(self, row, col, channels=None, colspan=1, yrange=None):
        plot = self.graphView.addPlot(
            # name=name,
            # title=title,
            row=row, col=col,
            colspan=colspan,
        )

        # tie all plot x-axes together
        if self.plots:
            plot.setXLink(self.plot0)

        # use automatic downsampling and clipping to reduce the
        # drawing load
        plot.setDownsampling(mode='peak')
        plot.setClipToView(True)
        plot.disableAutoRange(axis='x')
        plot.hideButtons()
        plot.showGrid(x=True, y=True, alpha=0.2)
        plot.addLegend()
        if yrange in [None, 'auto']:
            plot.enableAutoRange(axis='y')
        else:
            plot.setYRange(*yrange)

        plot._row = row
        plot._col = col
        plot._channels = {}
        plot._bad_data = []
        plot._is_trend = False
        plot._cursors = {}

        self.Crosshair.add_plot(plot)
        self.TCursors.add_plot(plot)

        if channels:
            for chan, kwargs in channels.items():
                kwargs = kwargs or {}
                self.add_channel_to_plot(plot, chan, **kwargs)

        self.plots.append(plot)
        return plot

    @property
    def plot0(self):
        return self.plots[0]

    def add_channel_to_plot(self, plot, channel, color=None, scale=1, offset=0):
        """Add a channel to an existing plot

        `color`, `scale` and `offset` are applied to the channel trace
        in the plot.  `color` can be a single letter color spec ('b',
        'r', etc.), an integer, or an [r,g,b] list.  See the following
        for more info:

          http://www.pyqtgraph.org/documentation/functions.html#pyqtgraph.mkColor

        """
        if channel in plot._channels:
            return

        label = channel

        if color:
            color = pg.mkColor(color)
            self.colors[channel] = color
        elif channel in self.colors:
            color = self.colors[channel]
        else:
            color = pg.mkColor(np.random.randint(20))
            self.colors[channel] = color
        mmc = copy.copy(color)
        mmc.setAlpha(50)

        # create a transform for the data
        if scale != 1 or offset != 0:
            def transform(data):
                return data*scale + offset
            if scale != 1:
                label += ' *{}'.format(scale)
            if offset != 0:
                label += ' +{}'.format(offset)
        else:
            def transform(data):
                return data

        cs = {}
        cs['y'] = plot.plot([0, 0], pen=color, name=label)
        cs['min'] = pg.PlotDataItem([0, 0], pen=mmc)
        cs['max'] = pg.PlotDataItem([0, 0], pen=mmc)
        # FIXME: HACK for bug in old pyqtgraph 0.9
        try:
            cs['fill'] = pg.FillBetweenItem(cs['min'], cs['max'], brush=mmc)
        except:
            cs['fill'] = None

        cs['transform'] = transform

        plot._channels[channel] = cs

        self.data.add_channel(channel)

    def remove_channel_from_plot(self, plot, channel):
        if channel not in plot._channels:
            return
        for curve in plot._curves[channel].values():
            plot.removeItem(curve)
        del plot._channels[channel]
        self.data.remove_channel(channel)

    def plots4chan(self, channel):
        plots = []
        for plot in self.plots:
            if channel in plot._channels:
                plots.append(plot)
        return plots

    ##########

    def start(self, lookback=None):
        """Start online mode

        """
        logging.info('START')
        self.online = True
        self.startstopButton.setIcon(self.style().standardIcon(QStyle.SP_MediaStop))
        self.startstopButton.setText("Stop")
        self.offlineTab.setEnabled(False)
        self.crosshairGroup.setChecked(False)
        self.crosshairGroup.setEnabled(False)
        if lookback:
            self.data.set_lookback(lookback)
        self.data.online(start_end=self.start_end)
        # FIXME: how do we not hard code the stride here
        self.setXRange(-self.data.lookback, 1/16)

    @pyqtSlot()
    def done(self):
        self.startstopButton.setIcon(self.style().standardIcon(QStyle.SP_MediaPlay))
        self.startstopButton.setText("Online")
        self.offlineTab.setEnabled(True)
        self.crosshairGroup.setEnabled(True)
        self.online = False

    def stop(self, message=None):
        """Stop online mode

        """
        logging.info('STOP')
        self.data.terminate()

    @pyqtSlot()
    def startstop(self):
        """Toggle online mode

        """
        if self.online:
            self.stop()
        else:
            self.start()

    def data_update(self, start, end, trend=None):
        span = end - start
        if trend is None:
            if span > TREND_THRESHOLD_M:
                trend = 'min'
            elif span > TREND_THRESHOLD_S:
                trend = 'sec'
            else:
                trend = 'raw'
        self.data.update((start, end), trend)

    @pyqtSlot()
    def fetch(self, start_end=None):
        """Fetch data offline

        """
        if not start_end:
            start = int(self.offlineStartGPS.text())
            end = int(self.offlineStopGPS.text())
        else:
            start, end = start_end
        logging.info('FETCH {}'.format((start, end)))
        self.toggle_trigger(False)
        self.triggerGroup.setChecked(False)
        self.data.reset()
        self.set_t0((end+start)/2)
        self.setXRange(start-self.t0, end-self.t0)
        self.data_update(start, end)
        self.updateGPS()

    @pyqtSlot()
    def update_range(self):
        (xmin, xmax), (ymin, ymax) = self.plot0.viewRange()
        start = self.t0 + xmin
        end = self.t0 + xmax
        if self.online:
            lookback = np.abs(xmin)
            logging.debug('update lookback: {}'.format(lookback))
            # FIXME: should be more rational way to do these two
            # things.  we have to check to make sure the lookback is
            # increasing to do the updates, since the extendleft
            # operation will fail as the append moves the data start
            # gps forward.  should be better way to do this.
            oldlb = self.data.lookback
            self.data.set_lookback(lookback)
            if lookback > oldlb:
                self.data_update(start, self.t0, 'raw')
        elif self.updateOnRange.isChecked():
            logging.debug('update range: {}'.format((start, end)))
            self.data_update(start, end)
        self.updateGPS()

    ##########
    # PLOTTING

    @pyqtSlot('PyQt_PyObject')
    def update_plots(self, recv):
        logging.log(5, "UPDATE PLOT")

        data, trend = recv
        is_trend = trend not in ['raw']

        if not data:
            self.clear_plots()
            return

        trigger = None
        if self.online:
            if self.trigger.active:
                trigger = self.trigger.check(data)
                if not trigger:
                    return
                self.update_trigger_time(trigger)
                self.set_t0(trigger)

            else:
                self.set_t0(data.gps_end)

        for plot in self.plots:
            # remove any previous bad data blocks
            for bdr in plot._bad_data:
                plot.removeItem(bdr)

            # add/remove trend min/max
            if is_trend and not plot._is_trend:
                for chan, curve in plot._channels.items():
                    plot.addItem(curve['min'])
                    plot.addItem(curve['max'])
                    if curve['fill']:
                        plot.addItem(curve['fill'])
                plot._is_trend = True
            elif not is_trend and plot._is_trend:
                for chan, curve in plot._channels.items():
                    plot.removeItem(curve['min'])
                    plot.removeItem(curve['max'])
                    if curve['fill']:
                        plot.removeItem(curve['fill'])
                plot._is_trend = False

            bad_data = None

            for chan, curve in plot._channels.items():
                cd = data[chan]
                t = cd.tarray - self.t0
                if cd.is_trend:
                    y = curve['transform'](cd.data['mean'])
                    plot._channels[chan]['min'].setData(
                        x=t,
                        y=curve['transform'](cd.data['min']),
                        connect="finite",
                    )
                    plot._channels[chan]['max'].setData(
                        x=t,
                        y=curve['transform'](cd.data['max']),
                        connect="finite",
                    )
                else:
                    y = curve['transform'](cd.data['raw'])

                curve['y'].setData(x=t, y=y, connect="finite")

                if np.any(np.isnan(y)):
                    if bad_data is None:
                        bad_data = np.isnan(y)
                    else:
                        bad_data = np.logical_or(np.isnan(y), bad_data, bad_data)

            # FIXME: the following slows this function down
            # consideraly.  should figure out a way to speed it up.
            # For now exclude from online plotting so it at least
            # doesn't slow that down.  Could just forget this
            # cleverness entirely and just leave the data gaps
            # there...
            if bad_data is not None and not self.online:
                # FIXME: feels like there should be a more
                # straightforward way to extract the bad data regions
                nani = np.where(bad_data)[0]
                bi = np.where(np.diff(nani) > 1)[0]
                bstart = np.append(nani[0], nani[bi + 1])
                bend = np.append(nani[bi], nani[-1])
                for bd in zip(bstart, bend):
                    bdt = [t[i] for i in bd]
                    bdr = pg.LinearRegionItem(
                        values=bdt,
                        brush=(250, 0, 0, 30),
                        #pen=pg.mkPen(None),
                        movable=False,
                    )
                    # FIXME, HACK: this is because for some reason the
                    # LinearRegionItem init does not accept the 'pen'
                    # kwarg, even though the documenatation says it
                    # should
                    for line in bdr.lines:
                        line.setPen(pg.mkPen(None))
                    plot.addItem(bdr)
                    plot._bad_data.append(bdr)

        if trigger and self.trigger.single:
            self.stop()

    def clear_plots(self):
        for plot in self.plots:
            for curves in plot._channels.values():
                curves['y'].setData(np.array([0, 0]))
                curves['min'].setData(np.array([0, 0]))
                curves['max'].setData(np.array([0, 0]))

    def set_t0(self, t0):
        self.t0 = t0
        label = 'seconds from {:0.4f} [{}]'.format(
            self.t0, gpstime.fromgps(t0).astimezone(DATETIME_TZ).strftime(DATETIME_FMT))
        self.referenceTimeLabel.setText(label)
        self.updateGPS()
        # FIXME: just use change event on above
        self.t0ChangeEvent.emit()

    def setXRange(self, start, end):
        logging.debug('RANGE: {} {}'.format(start, end))
        self.plot0.sigXRangeChanged.disconnect()
        self.plot0.setXRange(start, end, padding=0, update=False)
        self.plot0.sigXRangeChanged.connect(self.update_range)

    def updateGPS(self):
        (xmin, xmax), (ymin, ymax) = self.plot0.viewRange()
        start = int(self.t0 + xmin)
        end = int(np.ceil(self.t0 + xmax))
        self.offlineStartGPS.setText(str(start))
        self.offlineStopGPS.setText(str(end))

    ##########
    # TRIGGER

    def set_trigger(self, channel):
        assert channel in self.data.channels + [None]
        if channel == self.trigger.channel:
            return
        if self.trigger.channel is not None:
            tplot = self.plots4chan(self.trigger.channel)[0]
        else:
            tplot = None
        if channel is not None:
            nplot = self.plots4chan(channel)[0]
        else:
            nplot = None
        self.trigger.channel = channel
        if nplot != tplot:
            if tplot:
                tplot.removeItem(self.trigger.line)
            if nplot:
                nplot.addItem(self.trigger.line, ignoreBounds=True)
                nplot.disableAutoRange(axis='y')
            return True
        else:
            return False

    @pyqtSlot()
    def update_trigger_level(self):
        self.triggerLevel.setText('{:g}'.format(self.trigger.level))

    @pyqtSlot()
    def set_trigger_level(self):
        value = float(self.triggerLevel.text())
        self.trigger.set_level(value)

    @pyqtSlot()
    def reset_trigger_level(self):
        chan = self.trigger.channel
        if self.data['raw'] and chan in self.data['raw']:
            y = self.data['raw'][chan].data['raw']
            yn = y[np.where(np.invert(np.isnan(y)))[0]]
            value = np.mean(yn)
        else:
            value = 0
        self.trigger.set_level(value)

    def update_trigger_time(self, time):
        self.triggerTime.setText('{:14.3f}'.format(time))

    @pyqtSlot()
    def update_trigger_channel(self):
        chan = str(self.triggerSelect.currentText())
        if self.set_trigger(chan):
            self.reset_trigger_level()
        logging.info("trigger set: {}".format(chan))

    @pyqtSlot()
    def toggle_trigger(self, checked=None):
        if not self.trigger:
            return
        if checked is None:
            checked = self.triggerGroup.isChecked()
        if checked:
            chan = str(self.triggerSelect.currentText())
            self.set_trigger(chan)
            self.reset_trigger_level()
            logging.info("trigger enabled")
        else:
            self.set_trigger(None)
            logging.info("trigger disabled")

    ##########
    # CURSORS

    @pyqtSlot()
    def toggle_crosshair(self, checked=None):
        if checked is None:
            checked = self.crosshairGroup.isChecked()
        if checked:
            self.Crosshair.enable()
            # self.graphView.scene().sigMouseMoved.connect(self.Crosshair.update)
            self.crosshair_proxy = pg.SignalProxy(
                self.graphView.scene().sigMouseMoved,
                rateLimit=60,
                slot=self.Crosshair.update)
            logging.info("crosshair enabled")
        else:
            self.Crosshair.disable()
            self.crosshair_proxy = None
            logging.info("crosshair disabled")

    @pyqtSlot()
    def toggle_t_cursors(self, checked=None):
        if checked is None:
            checked = self.cursorTGroup.isChecked()
        if checked:
            self.TCursors.enable()
            logging.info("T-cursor enabled")
        else:
            self.TCursors.disable()
            logging.info("T-cursor disabled")

    def get_selected_y_cursors_plot(self):
        rc = self.cursorYPlot.currentText()
        r, c = [int(n) for n in rc.split(',')]
        return self.graphView.getItem(r, c)

    @pyqtSlot()
    def set_y_cursors_to_selected(self):
        plot = self.get_selected_y_cursors_plot()
        self.YCursors.set_plot(plot)

    @pyqtSlot()
    def toggle_y_cursors(self, checked=None):
        if checked is None:
            checked = self.cursorYGroup.isChecked()
        if checked:
            self.set_y_cursors_to_selected()
            logging.info("Y-cursor enabled")
        else:
            self.YCursors.set_plot(None)
            logging.info("Y-cursor disabled")
