from __future__ import division
import os
import time
import numpy as np
from contextlib import closing
from PyQt4.QtCore import QThread, QMutex, pyqtSignal
import logging

import nds2
from gpstime import gpsnow

##########

TREND_CTYPES = [
    nds2.channel.CHANNEL_TYPE_STREND,
    nds2.channel.CHANNEL_TYPE_MTREND,
]

try:
    HOSTPORT = os.getenv('NDSSERVER').split(',')[0].split(':')
except AttributeError:
    raise SystemExit("NDSSERVER environment variable not specified.")
HOST = HOSTPORT[0]
try:
    PORT = int(HOSTPORT[1])
except IndexError:
    PORT = 31200

##########

class FakeBuffer:
    def __init__(self, name, start, stride):
        class Channel:
            pass
        self.channel = Channel()
        ct = name.split(',')
        if len(ct) > 1:
            if ct[1] == 's-trend':
                ctype = nds2.channel.CHANNEL_TYPE_STREND
                sample_rate = 1
            elif ct[1] == 'm-trend':
                ctype = nds2.channel.CHANNEL_TYPE_MTREND
                sample_rate = 1/60
        else:
            ctype = nds2.channel.CHANNEL_TYPE_RAW
            sample_rate = 2**14
        self.channel.name = ct[0]
        self.channel.channel_type = ctype
        self.channel.sample_rate = sample_rate
        self.gps_seconds = int(start)
        self.gps_nanoseconds = int((start % 1) * 1e9)
        def sampler(s):
            signal = 10*np.sin(np.arange(s)*2*np.pi/16)
            noise = np.random.exponential(3, s)
            # noise = np.random.random(s)
            return signal + noise
        self.data = sampler(int(sample_rate*stride))
        nbd = np.random.randint(0, len(self.data)/2)
        if (self.gps_seconds % 10) == 0 \
           and self.gps_nanoseconds == 0:
            self.data[nbd:nbd+200] *= np.nan


class FakeConnection:
    def set_parameter(self, *args):
        return

    def get_protocol(self):
        return 1

    def fetch(self, start, end, channels):
        stride = end - start
        return [FakeBuffer(chan, start, stride) for chan in channels]

    def iterate(self, stride, channels):
        if stride == -1:
            stride = 1./16
        start = int(gpsnow())
        while True:
            yield [FakeBuffer(chan, start, stride) for chan in channels]
            start += stride
            sleep = start - gpsnow()
            if sleep > 0:
                time.sleep(sleep)

    def close(self):
        return

##########

def get_connection():
    logging.debug("NDS connect: {}:{}".format(HOST, PORT))
    if HOST == 'fake':
        conn = FakeConnection()
    else:
        conn = nds2.connection(HOST, PORT)
    conn.set_parameter('GAP_HANDLER', 'STATIC_HANDLER_NAN')
    # conn.set_parameter('ITERATE_USE_GAP_HANDLERS', 'false')
    return conn


def iterate(channels, start_end=None):
    args = []
    if start_end:
        args += list(start_end)
    # FIXME: nds2 0.16 has the following:
    # args += [nds2.connection.FAST_STRIDE]
    # args += [channels]
    with closing(get_connection()) as conn:
        if conn.get_protocol() == 1:
            stride = -1
        else:
            stride = 1
        args += [stride]
        args += [channels]
        for bufs in conn.iterate(*args):
            yield bufs


def fetch(channels, start_end):
    args = list(start_end) + [channels]
    with closing(get_connection()) as conn:
        return conn.fetch(*args)


def parse_channel(channel):
    ctype = nds2.channel.channel_type_to_string(channel.channel_type)
    if ctype in ['s-trend', 'm-trend']:
        name, mod = channel.name.split('.')
    else:
        name = channel.name
        mod = 'raw'
    return name, mod, ctype

##################################################

class NDSThread(QThread):
    new_data = pyqtSignal('PyQt_PyObject')
    done = pyqtSignal('PyQt_PyObject')

    _type_map = {'raw': None, 'sec': 's', 'min': 'm'}

    def __init__(self, method, tid, trend, channels, start_end=None):
        assert method in ['fetch', 'iterate']
        if method == 'fetch':
            assert start_end is not None
        super(NDSThread, self).__init__()
        self.method = method
        self.tid = tid
        self.trend = trend
        self.channels = []
        for chan in channels:
            if trend == 'raw':
                self.channels.append(chan)
            else:
                for m in ['mean', 'min', 'max']:
                    t = self._type_map[self.trend]
                    self.channels.append('{}.{},{}-trend'.format(chan, m, t))
        if start_end:
            start = int(start_end[0])
            end = int(np.ceil(start_end[1]))
            if trend == 'min':
                start -= start % 60
                end -= (end % 60)
            self.start_end = (start, end)
        else:
            self.start_end = None
        self._run_lock = QMutex()
        self._running = True

    def emit_data(self, bufs):
        self.new_data.emit((self.tid, self.trend, bufs))

    def emit_done(self, error=None):
        self.done.emit((self.tid, self.trend, error))

    @property
    def running(self):
        try:
            self._run_lock.lock()
        # FIXME: python3
        #else:
            return self._running
        finally:
            self._run_lock.unlock()

    def run(self):
        logging.log(5, "NDS {} {} {}".format(self.method, self.start_end, self.channels))
        error = None

        if self.method == 'fetch':
            try:
                bufs = fetch(self.channels, self.start_end)
                self.emit_data(bufs)
                logging.log(5, "NDS fetched: {} {}".format(bufs[0].gps_seconds, bufs[0].gps_nanoseconds))
            except RuntimeError as e:
                error = str(e).split('\n')[0]

        elif self.method == 'iterate':
            try:
                for bufs in iterate(self.channels, self.start_end):
                    if not self.running:
                        break
                    self.emit_data(bufs)
                    logging.log(5, "NDS iterate: {} {}".format(bufs[0].gps_seconds, bufs[0].gps_nanoseconds))
                    if not self.running:
                        break
            except RuntimeError as e:
                error = str(e).split('\n')[0]

        self.emit_done(error)

    def stop(self):
        self._run_lock.lock()
        self._running = False
        self._run_lock.unlock()
